from setuptools import setup

setup(
    name='UMNN',
    version='1.7',
    packages=['UMNN'],
    url='',
    license='MIT License',
    author='awehenkel',
    author_email='antoine.wehenkel@gmail.com',
    description='Unconstrained Monotonic Neural Networks and Autoregressive Monotonic Flows on cpu/cuda/mps.'
)
